# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import logging
import os
from collections import defaultdict
from datetime import datetime
from enum import Enum, auto

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from xdg import xdg_cache_home, xdg_data_home

from .tasks import Task, TaskList, TaskStatus

CREDENTIALS_FILE = "credentials.json"
SCOPES = ["https://www.googleapis.com/auth/tasks"]


# https://googleapis.github.io/google-api-python-client/docs/dyn/tasks_v1.html
class GoogleApiService:
    def __init__(
        self,
        user: str,
        completed_after: datetime | None,
        completed_before: datetime | None,
        task_status: TaskStatus,
    ):
        self.user = user
        self.completed_after = completed_after
        self.completed_before = completed_before
        self.task_status = TaskStatus(task_status) if task_status else None
        self._service = None

    def tasks(self):
        return self._get_service().tasks()

    def task_lists(self):
        return self._get_service().tasklists()

    def new_batch_http_request(self):
        return self._get_service().new_batch_http_request()

    async def reconcile(
        self, old_task_lists: list[TaskList], new_task_lists: list[TaskList]
    ):
        """
        Reconciles differences between new and old task lists.

        After a user modifies state containing all tasklists with their tasks,
        it's needed to reconcile resulting differences. The logic for task
        lists, tasks and subtasks is somewhat similar. At first all the old
        task lists are marked to be removed. Then for every new task lists if
        it's ID was set to be removed then it's reconciled (updated) instead.
        Otherwise such task list is marked to be added. In the end the order of
        items is restored. Items that have the same old and new state are
        skipped. All the items are matched based on title.
        """

        def gen_tasklist_ops():
            task_list_to_op = {}
            for task_list in old_task_lists:
                task_list_to_op[task_list.title] = (ReconcileOp.DELETE, task_list)

            for task_list in new_task_lists:
                if task_list.title in task_list_to_op:
                    task_list_to_op[task_list.title] = (
                        ReconcileOp.UPDATE,
                        task_list_to_op[task_list.title][1],
                        task_list,
                    )
                else:
                    task_list_to_op[task_list.title] = (ReconcileOp.INSERT, task_list)

            return list(task_list_to_op.values())

        async def apply_task_list_op(op):
            match op:
                case (ReconcileOp.DELETE, task_list):
                    self.task_lists().delete(tasklist=task_list.id).execute()
                    logging.info(f"Deleted Task List {task_list.title}")

                case (ReconcileOp.INSERT, task_list):
                    response = (
                        self.task_lists().insert(body=task_list.toRequest()).execute()
                    )
                    reconcile_tasks(response["id"], [], task_list.tasks)
                    logging.info(f"Inserted Task List {task_list.title}")

                case (ReconcileOp.UPDATE, old_task_list, new_task_list):
                    if old_task_list != new_task_list:
                        reconcile_tasks(
                            old_task_list.id, old_task_list.tasks, new_task_list.tasks
                        )
                        logging.info(f"Updated Task List {old_task_list.title}")

        def reconcile_tasks(task_list_id, old_tasks, new_tasks, parent_task_id=""):
            updated_tasks = apply_task_ops(
                gen_task_ops(old_tasks, new_tasks),
                new_tasks,
                task_list_id,
            )

            completed_tasks = []
            incompleted_tasks = []
            for task in updated_tasks:
                if task.completed():
                    completed_tasks.append(task)
                else:
                    incompleted_tasks.append(task)
            fix_task_order(
                task_list_id,
                list(incompleted_tasks),
                parent_task_id,
            )
            fix_task_order(
                task_list_id,
                list(completed_tasks),
                parent_task_id,
            )
            return updated_tasks

        def apply_task_ops(ops, new_tasks, task_list_id):
            def delete_callback(task_title):
                def callback(request_id, response, exception):
                    del request_id, response
                    if exception:
                        logging.error(f"Failed to delete task {task_title}")
                    else:
                        logging.info(f"Deleted Task {task_title}")

                return callback

            def insert_callback(task, idx):
                def callback(_, response, exception):
                    if exception:
                        logging.error(f"Failed to insert task {task.title}")

                    task_id = response["id"]
                    new_tasks[idx].id = task_id  # Needed for fix_task_order

                    updated_subtasks = reconcile_tasks(
                        task_list_id, [], task.subtasks, task_id
                    )
                    new_tasks[idx].subtasks = updated_subtasks
                    logging.info(f"Inserted Task {task.title}")

                return callback

            def update_callback(old_task, new_task, idx):
                def callback(request_id, response, exception):
                    del request_id, response
                    if exception:
                        logging.error(f"Failed to update Task {old_task.title}")
                        return

                    updated_subtasks = reconcile_tasks(
                        task_list_id, old_task.subtasks, new_task.subtasks, old_task.id
                    )
                    new_tasks[idx].subtasks = updated_subtasks
                    logging.info(f"Updated Task {old_task.title}")

                return callback

            batched_request = self.new_batch_http_request()
            for op in ops:
                match op:
                    case (ReconcileOp.DELETE, task):
                        batched_request.add(
                            self.tasks().delete(tasklist=task_list_id, task=task.id),
                            delete_callback(task.title),
                        )
                    case (ReconcileOp.INSERT, task, idx):
                        batched_request.add(
                            self.tasks().insert(
                                tasklist=task_list_id, body=task.toRequest()
                            ),
                            insert_callback(task, idx),
                        )
                    case (ReconcileOp.UPDATE, old_task, new_task, idx):
                        new_tasks[idx].id = old_task.id  # Needed for fix_task_order
                        if old_task != new_task:
                            batched_request.add(
                                self.tasks().patch(
                                    tasklist=task_list_id,
                                    task=old_task.id,
                                    body=new_task.toRequest(),
                                ),
                                update_callback(old_task, new_task, idx),
                            )
            batched_request.execute()

            return new_tasks

        def gen_task_ops(old_tasks: list[Task], new_tasks: list[Task]):
            task_to_op = {}
            for task in old_tasks:
                task_to_op[task.title] = (ReconcileOp.DELETE, task)

            for i, task in enumerate(new_tasks):
                if task.title in task_to_op:
                    task_to_op[task.title] = (
                        ReconcileOp.UPDATE,
                        task_to_op[task.title][1],
                        task,
                        i,
                    )
                else:
                    task_to_op[task.title] = (ReconcileOp.INSERT, task, i)

            return list(task_to_op.values())

        # The move requests can't be sent in parallel as there must not be
        # two values pointing to the same predecessor.
        def fix_task_order(task_list_id, new_tasks, parent_task_id=""):
            for i, task in enumerate(new_tasks):
                previous_task = new_tasks[i - 1] if i > 0 else None
                previous_task_id = previous_task.id if previous_task else ""

                self.tasks().move(
                    tasklist=task_list_id,
                    task=task.id,
                    parent=parent_task_id,
                    previous=previous_task_id,
                ).execute()

                previous_task_title = previous_task.title if previous_task else "NONE"
                logging.info(
                    f"Moved task {task.title} after {previous_task_title} (parent: {parent_task_id})"
                )

        async_tasks = []
        for op in gen_tasklist_ops():
            async_tasks.append(asyncio.create_task(apply_task_list_op(op)))
        await asyncio.gather(*async_tasks)

    def fetch_task_lists(self) -> list[TaskList]:
        """
        Fetches all tasks from the server.

        At first the function fetches up to 100 task lists. Then it fetches all
        tasks for these task lists that are either completed at most 30 days ago
        or are still pending completion.
        """
        id_to_task_list = {}
        task_id_to_subtasks = defaultdict(list)

        def create_request_with_callback(task_list_id, completed):
            def fetch_tasks_request(task_list_id, completed, next_page_token=""):
                completed_max = ""
                completed_min = ""
                if completed:
                    if self.completed_before:
                        completed_max = self.completed_before.isoformat()
                    if self.completed_after:
                        completed_min = self.completed_after.isoformat()

                return self.tasks().list(
                    completedMax=completed_max,
                    completedMin=completed_min,
                    maxResults=100,
                    pageToken=next_page_token,
                    showCompleted=completed,
                    showHidden=completed,
                    tasklist=task_list_id,
                )

            def callback(_, response, exception):
                if exception:
                    logging.error(
                        f"Error on fetching Tasks from Task List {task_list_id}: {exception}"
                    )
                    return

                fetched_tasks = response.get("items", [])
                next_page_token = response.get("nextPageToken", "")
                while next_page_token:
                    response = fetch_tasks_request(
                        task_list_id, completed, next_page_token
                    ).execute()
                    fetched_tasks += response.get("items", [])
                    next_page_token = response.get("nextPageToken", "")

                for fetched_task in fetched_tasks:
                    task = Task(
                        fetched_task["id"],
                        fetched_task["title"].strip(),
                        fetched_task.get("notes", ""),
                        int(fetched_task["position"]),
                        TaskStatus(fetched_task.get("status", "unknown")),
                        [],
                    )

                    # If a task has a parent then it's definitely a subtask
                    # Subtask's parent might be incompleted so appending it
                    # to it must be deferred.
                    parent = fetched_task.get("parent", "")
                    if parent:
                        task_id_to_subtasks[parent].append(task)
                    else:
                        id_to_task_list[task_list_id].tasks.append(task)

            return fetch_tasks_request(task_list_id, completed), callback

        task_lists = self.task_lists().list(maxResults=100).execute().get("items", [])

        batched_request = self.new_batch_http_request()
        for task_list in task_lists:
            id = task_list["id"]
            id_to_task_list[id] = TaskList(id, task_list["title"], [])

            if not self.task_status or self.task_status == TaskStatus.PENDING:
                batched_request.add(*create_request_with_callback(id, False))
            if not self.task_status or self.task_status == TaskStatus.COMPLETED:
                batched_request.add(*create_request_with_callback(id, True))
        batched_request.execute()

        task_lists = list(id_to_task_list.values())
        task_lists.sort(key=lambda tl: tl.title)
        for task_list in task_lists:
            for task in task_list.tasks:
                task.subtasks = task_id_to_subtasks.get(task.id, [])
                task.subtasks.sort(key=lambda t: t.position)
            task_list.tasks.sort(key=lambda t: t.position)

        return task_lists

    # https://developers.google.com/tasks/quickstart/python#step_2_configure_the_sample
    def get_credentials(self) -> Credentials:
        """
        Read credentials from selected user configuration.

        This function will try to read existing token from
        $XDG_CACHE_HOME/gtasks-md directory for the selected user. If file with
        the token doesn't exist, it will try creating a new one after reading
        credentials from $XDG_DATA_HOME/gtasks-md. If there are no credentials
        the process will simply fail.
        """
        creds = None
        config_dir = f"{xdg_data_home()}/gtasks-md/{self.user}"
        credentials_file = f"{config_dir}/{CREDENTIALS_FILE}"
        cache_dir = f"{xdg_cache_home()}/gtasks-md/{self.user}"
        token_file = f"{cache_dir}/token.json"
        # The file token.json stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first
        # time.
        if os.path.exists(token_file):
            creds = Credentials.from_authorized_user_file(token_file, SCOPES)
        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    credentials_file, SCOPES
                )
                creds = flow.run_local_server(port=0)
            # Save the credentials for the next run
            with open(token_file, "w+") as token:
                token.write(creds.to_json())

        return creds

    def save_credentials(self, credentials: str):
        """Save credentials to selected user config directory."""
        config_dir = f"{xdg_data_home()}/gtasks-md/{self.user}"

        with open(f"{config_dir}/{CREDENTIALS_FILE}", "w+") as dest_file:
            dest_file.write(credentials)

    def _get_service(self):
        if not self._service:
            self._service = build(
                "tasks",
                "v1",
                credentials=self.get_credentials(),
                cache_discovery=False,
                static_discovery=True,
            )
        return self._service


class ReconcileOp(Enum):
    INSERT = auto()
    DELETE = auto()
    UPDATE = auto()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import subprocess
import tempfile


class Editor:
    """
    Handles all editor-related functionality.
    """

    def __init__(self, editor):
        if editor:
            self.editor = editor
        elif "VISUAL" in os.environ:
            self.editor = os.environ["VISUAL"]
        elif "EDITOR" in os.environ:
            self.editor = os.environ["EDITOR"]
        else:
            self.editor = "vim"

    def edit(self, text: str) -> str:
        """
        Edit given markdown text.

        A backup of the input will be stored in the $XDG_CACHE_HOME/gtasks-md
        directory. Next, the text will be stored in a temporary file that will
        be edited with provided editor. In case the editor is missing, the
        program will try the following options: $VISUAL, $EDITOR and "vim".
        """

        tmp_file = ""
        with tempfile.NamedTemporaryFile(suffix=".md", delete=False) as tmp:
            tmp.write(str.encode(text))
            tmp.flush()
            if subprocess.call([self.editor, tmp.name]) != 0:
                exit(1)
            tmp_file = tmp.name

        out = ""
        with open(tmp_file, "r") as output:
            out = output.read()

        os.remove(tmp_file)
        return out



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pandoc
from pandoc import types

from .tasks import Task, TaskList, TaskStatus

# https://github.com/jgm/pandoc-types/blob/master/src/Text/Pandoc/Definition.hs
Decimal = types.Decimal  # type: ignore
Header = types.Header  # type: ignore
Meta = types.Meta  # type: ignore
OrderedList = types.OrderedList  # type: ignore
Pandoc = types.Pandoc  # type: ignore
Para = types.Para  # type: ignore
Period = types.Period  # type: ignore
Plain = types.Plain  # type: ignore
SoftBreak = types.SoftBreak  # type: ignore
Space = types.Space  # type: ignore
Str = types.Str  # type: ignore

EMPTY_ATTRS = ("", [], [])
ORDERED_FIRST_ELEM = (1, Decimal(), Period())


def task_lists_to_markdown(task_lists: list[TaskList]) -> str:
    """Parses Task Lists to a Pandoc markdown"""

    def text_to_pandoc(text: str):
        elems = []
        for word in text.split():
            elems.append(Str(word))
            elems.append(Space())
        return elems[:-1]

    def tasks_to_pandoc(tasks: list[Task]):
        pandoc_tasks = []
        has_notes = any(t.note for t in tasks)
        for task in tasks:
            pandoc_tasks.append(task_to_pandoc(task, has_notes))
        return pandoc_tasks

    def task_to_pandoc(task: Task, parent_contains_notes: bool):
        pandocTask = []

        task_sign = "☒" if task.completed() else "☐"
        task_title = [Str(task_sign), Space()] + text_to_pandoc(task.title)

        if parent_contains_notes:
            pandocTask.append(Para(task_title))
            match pandoc.read(task.note):
                case Pandoc(_, [*note]):
                    pandocTask += note
                case Pandoc(_, []):
                    pass
                case _:
                    raise SyntaxError(f"Could not parse Task note:\n{task.note}")
        else:
            pandocTask.append(Plain(task_title))

        if task.subtasks:
            subtasks = []
            parent_contains_notes = any(st.note for st in task.subtasks)
            for subtask in task.subtasks:
                subtasks.append(task_to_pandoc(subtask, parent_contains_notes))

            pandocTask.append(OrderedList(ORDERED_FIRST_ELEM, subtasks))

        return pandocTask

    content = [
        Header(1, EMPTY_ATTRS, [Str("Google"), Space(), Str("Tasks")]),
    ]

    for task_list in task_lists:
        content.append(Header(2, EMPTY_ATTRS, text_to_pandoc(task_list.title)))
        content.append(
            OrderedList(ORDERED_FIRST_ELEM, tasks_to_pandoc(task_list.tasks))
        )

    return pandoc.write(Pandoc(Meta({}), content))


def markdown_to_task_lists(text: str) -> list[TaskList]:
    """Parses Pandoc markdown to Task Lists"""

    def parse_task_lists(items, idx):
        if idx >= len(items):
            return []

        match items[idx]:
            case Header(1, _, _):
                return parse_task_lists(items, idx + 1)
            case Header(2, _, hd):
                task_list = TaskList("", pandoc.write(hd).strip(), [])

                if idx + 1 < len(items):
                    match items[idx + 1]:
                        case OrderedList(_, tasks):
                            task_list.tasks = parse_tasks(tasks)
                            return [task_list] + parse_task_lists(items, idx + 2)
                        case _:
                            return [task_list] + parse_task_lists(items, idx + 1)
                else:
                    return [task_list]
            case _:
                raise SyntaxError(f"Unexpected item while parsing: {items[idx]}")

    def parse_tasks(tasks):
        parsed_tasks = []

        for i, task in enumerate(tasks):
            parsed_tasks.append(parse_task(task, i))

        return parsed_tasks

    def parse_task(task, taskNo):
        def match_status(str: Str) -> TaskStatus:
            match str:
                case Str("☐"):
                    return TaskStatus.PENDING
                case Str("☒"):
                    return TaskStatus.COMPLETED
                case _:
                    return TaskStatus.UNKNOWN

        name = ""
        status = TaskStatus.UNKNOWN
        match task[0]:
            case Plain(txt) | Para(txt):
                status = match_status(txt[0])
                if status == TaskStatus.UNKNOWN:
                    name = pandoc.write(Plain(txt))
                else:
                    name = pandoc.write(Plain(txt[2:]))
            case _:
                raise SyntaxError(f"Expected Task status and title, got {task[0]}")

        note = ""
        subtasks = []
        match task[-1]:
            case OrderedList(_, subtasks):
                note = pandoc.write(
                    Pandoc(Meta({}), task[1:-1]), options=["--wrap=none"]
                )
                subtasks = parse_tasks(subtasks)
            case _:
                note = pandoc.write(Pandoc(Meta({}), task[1:]), options=["--wrap=none"])

        return Task("", name.strip(), note.strip(), taskNo, status, subtasks)

    match pandoc.read(text):
        case Pandoc(_, items):
            return parse_task_lists(items, 0)
        case _:
            raise SyntaxError("Expected Pandoc markdown representation.")



import unittest
from inspect import cleandoc

from .pandoc import markdown_to_task_lists, task_lists_to_markdown
from .tasks import Task, TaskList, TaskStatus


class TestPandocConversion(unittest.TestCase):
    def test_header_only(self):
        markdown = """
        # Google Tasks
        """

        self.assertEqualAfterParsing([], markdown)

    def test_task_list(self):
        task_list = create_task_list("Task List 1")
        markdown = """
        # Google Tasks

        ## Task List 1
        """

        self.assertEqualAfterParsing([task_list], markdown)

    def test_task_list_with_tasks(self):
        task_list = create_task_list(
            "Task List 1",
            create_task("Task 1"),
            create_task("Task 2", status=TaskStatus.COMPLETED),
            create_task("Task 3", "Some note."),
        )
        markdown = """
        # Google Tasks

        ## Task List 1

        1.  [ ] Task 1

        2.  [x] Task 2

        3.  [ ] Task 3

            Some note.
        """

        self.assertEqualAfterParsing([task_list], markdown)

    def test_task_list_with_task_with_subtasks(self):
        task_list = create_task_list(
            "Task List 1",
            create_task(
                "Task 1",
                note="Some note",
                subtasks=[
                    create_task("Subtask 1"),
                    create_task("Subtask 2"),
                    create_task("Subtask 3"),
                ],
            ),
        )
        markdown = """
        # Google Tasks

        ## Task List 1

        1.  [ ] Task 1

            Some note

            1.  [ ] Subtask 1
            2.  [ ] Subtask 2
            3.  [ ] Subtask 3
        """

        self.assertEqualAfterParsing([task_list], markdown)

    def test_fail_to_parse_invalid_header(self):
        markdown = """
        # Google Tasks

        ## Task List 1

        ### Task 1
        """

        self.assertRaises(SyntaxError, markdown_to_task_lists, markdown)

    def test_fail_to_parse_unexpected_paragraph(self):
        markdown = """
        # Google Tasks

        ## Task List 1

        1.  [ ] Task 1
        2.  [x] Task 2

        Some paragraph.
        """

        self.assertRaises(SyntaxError, markdown_to_task_lists, markdown)

    def assertEqualAfterParsing(self, task_lists: list[TaskList], markdown: str):
        parsed_markdown = task_lists_to_markdown(task_lists)
        self.assertEqualMarkdown(markdown, parsed_markdown)
        parsed_task_lists = markdown_to_task_lists(parsed_markdown)
        self.assertEqual(task_lists, parsed_task_lists)

    def assertEqualMarkdown(self, text_1: str, text_2: str):
        self.assertEqual(cleandoc(text_1.strip()), cleandoc(text_2.strip()))


def create_task_list(name: str, *tasks) -> TaskList:
    return TaskList("", name, list(tasks))


def create_task(
    title: str,
    note: str = "",
    status: TaskStatus = TaskStatus.PENDING,
    subtasks: list[Task] = [],
) -> Task:
    return Task("", title, note, 0, status, subtasks)


if __name__ == "__main__":
    unittest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from dataclasses import dataclass
from enum import StrEnum


class TaskStatus(StrEnum):
    UNKNOWN = "unknown"
    PENDING = "needsAction"
    COMPLETED = "completed"

    @classmethod
    def _missing_(cls, value: str):
        for member in cls:
            if member.value.casefold() == value.casefold():
                return member
        return None


# https://developers.google.com/tasks/reference/rest/v1/tasks
@dataclass
class Task:
    """Task definition matching Google Task API"""

    id: str
    title: str
    note: str
    position: int
    status: TaskStatus
    subtasks: list[Task]

    def __eq__(self, other: Task) -> bool:
        return (
            self.title == other.title
            and self.note == other.note
            and self.status == other.status
            and len(self.subtasks) == len(other.subtasks)
            and not any(st != ot for (st, ot) in zip(self.subtasks, other.subtasks))
        )

    def __str__(self) -> str:
        return f"#{self.position}: {self.title} ({self.id}): {self.status}, {len(self.subtasks)} subtasks"

    def completed(self) -> bool:
        return self.status == TaskStatus.COMPLETED

    def toRequest(self) -> dict[str, str]:
        return {
            "id": self.id,
            "kind": "tasks#task",
            "notes": self.note,
            "status": self.status.value,
            "title": self.title,
        }


# https://developers.google.com/tasks/reference/rest/v1/tasklists
@dataclass
class TaskList:
    """Tasklist definition matching Google Task API"""

    id: str
    title: str
    tasks: list[Task]

    def __eq__(self, other: TaskList) -> bool:
        return (
            self.title == other.title
            and len(self.tasks) == len(other.tasks)
            and not any(st != ot for (st, ot) in zip(self.tasks, other.tasks))
        )

    def __str__(self) -> str:
        return f"{self.title} ({self.id}): {len(self.tasks)} tasks"

    def toRequest(self) -> dict[str, str]:
        return {
            "id": self.id,
            "kind": "tasks#taskList",
            "title": self.title,
        }



#!/usr/bin/env python3
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from os import path

from setuptools import setup

here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="gtasks-md",
    author="Michal Kielbowicz",
    author_email="gtasks-md@kielbowi.cz",
    url="https://github.com/google/gtasks-md",
    description="A tool to manage Google Tasks using a markdown document.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.11",
    install_requires=[
        "setuptools",
        "google-api-python-client",
        "google-auth-httplib2",
        "google-auth-oauthlib",
        "pandoc",
        "xdg",
    ],
    entry_points={"console_scripts": ["gtasks-md=app.__main__:main"]},
    packages=["app"],
    version="0.0.10",
    license="Apache License 2.0",
)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pathlib import Path

from xdg import xdg_cache_home


class Backup:
    """
    Handles all backup-related functionality.
    """

    def __init__(self, user):
        self.user = user

    def write_backup(self, text: str):
        cache_dir = f"{xdg_cache_home()}/gtasks-md/{self.user}"
        marker_file_path = Path(f"{cache_dir}/marker")

        marker_file_path.touch()
        with marker_file_path.open("r+") as marker_file:
            marker = marker_file.read()
            file_no = (int(marker) + 1 if marker else 0) % 10
            marker_file.seek(0)
            marker_file.write(str(file_no))
            marker_file.truncate()

            with open(f"{cache_dir}/{file_no}.bak.md", "w") as backup_file:
                backup_file.write(text)

    def discard_backup(self):
        cache_dir = f"{xdg_cache_home()}/gtasks-md/{self.user}"
        marker_file_path = Path(f"{cache_dir}/marker")

        if not marker_file_path.is_file():
            return None

        with marker_file_path.open("r+") as marker_file:
            marker = marker_file.read()
            file_no = ((int(marker) if marker else 0)) % 10
            marker_file.seek(0)
            marker_file.write(str(file_no - 1))
            marker_file.truncate()

            return f"{cache_dir}/{file_no}.bak.md"



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import asyncio
import datetime
import logging
import os
from datetime import timedelta

from xdg import xdg_cache_home, xdg_data_home

from .backup import Backup
from .editor import Editor
from .googleapi import GoogleApiService
from .pandoc import markdown_to_task_lists, task_lists_to_markdown


def main():
    args = parse_args()

    config_dir = f"{xdg_data_home()}/gtasks-md/{args.user}/"
    os.makedirs(os.path.dirname(config_dir), exist_ok=True)
    cache_dir = f"{xdg_cache_home()}/gtasks-md/{args.user}/"
    os.makedirs(os.path.dirname(cache_dir), exist_ok=True)

    logging.basicConfig(
        filename=f"{xdg_cache_home()}/gtasks-md/log.txt",
        format="%(asctime)s %(levelname)-8s %(message)s",
        encoding="utf-8",
        level=logging.INFO,
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    service = GoogleApiService(
        args.user, args.completed_after, args.completed_before, args.status
    )
    match args.subcommand:
        case "auth":
            auth(service, args.credentials_file)
        case "edit":
            editor = Editor(args.editor)
            backup = Backup(args.user)
            edit(service, editor, backup)
        case "reconcile":
            backup = Backup(args.user)
            reconcile(service, args.file_path, backup)
        case "rollback":
            backup = Backup(args.user)
            rollback(service, backup)
        case "view":
            view(service)
        case None:
            print("Please run one of the subcommands.")


def parse_args():
    def parse_date(date):
        return datetime.datetime.strptime(date, "%Y-%m-%d").astimezone()

    # https://stackoverflow.com/a/49977713
    parser = argparse.ArgumentParser(description="Google Tasks declarative management.")
    parser.add_argument(
        "--completed-after",
        dest="completed_after",
        default=(datetime.datetime.now() - timedelta(days=7)).astimezone(),
        help="Only show tasks completed after given date. The date must be in format YYYY-MM-DD. Defaults to one week ago.",
        type=parse_date,
    )
    parser.add_argument(
        "--completed-before",
        dest="completed_before",
        default=None,
        help="Only show tasks completed before given date. The date must be in format YYYY-MM-DD.",
        type=lambda d: parse_date(d) if d else None,
    )
    parser.add_argument(
        "--status",
        dest="status",
        default="",
        help="Task status. One of: needsAction, completed.",
        type=str.lower,
    )
    parser.add_argument(
        "--user",
        dest="user",
        default="default",
        help="Account for which the credentials are sourced. Should match desired Google account.",
        type=str,
    )

    subparsers = parser.add_subparsers(dest="subcommand")

    auth = subparsers.add_parser("auth", help="Authorize.")
    auth.add_argument(
        "credentials_file",
        help="Location of credential file.",
        type=str,
    )

    edit_parser = subparsers.add_parser("edit", help="Edit Google Tasks.")
    edit_parser.add_argument(
        "--editor",
        dest="editor",
        default="",
        help="Editor to be used for editing tasks. Defaults to $EDITOR and then $VISUAL and then vim.",
        type=str,
    )

    reconcile_parser = subparsers.add_parser(
        "reconcile", help="Patch Task Lists with an offline source."
    )
    reconcile_parser.add_argument(
        "file_path",
        help="Location of the source file.",
        type=str,
    )

    subparsers.add_parser("rollback", help="Rollback last change.")
    subparsers.add_parser("view", help="View Google Tasks.")

    return parser.parse_args()


def auth(service: GoogleApiService, file: str):
    with open(file, "r") as src_file:
        service.save_credentials(src_file.read())


def view(service: GoogleApiService):
    _, text = fetch_task_lists(service)
    print(text)


def edit(service: GoogleApiService, editor: Editor, backup: Backup):
    old_task_lists, old_text = fetch_task_lists(service)
    new_text = editor.edit(old_text)
    new_task_lists = markdown_to_task_lists(new_text)
    backup.write_backup(old_text)
    asyncio.run(service.reconcile(old_task_lists, new_task_lists))


def reconcile(service: GoogleApiService, file_path: str, backup: Backup | None = None):
    old_task_lists, old_text = fetch_task_lists(service)

    with open(file_path, "r") as source:
        new_text = source.read()
        new_task_lists = markdown_to_task_lists(new_text)
        if backup:
            backup.write_backup(old_text)
        asyncio.run(service.reconcile(old_task_lists, new_task_lists))


def rollback(service: GoogleApiService, backup: Backup):
    backup_file = backup.discard_backup()
    if backup_file:
        reconcile(service, backup_file, None)
    else:
        print("No backup found")


def fetch_task_lists(service: GoogleApiService):
    task_lists = service.fetch_task_lists()
    return task_lists, task_lists_to_markdown(task_lists)


if __name__ == "__main__":
    main()
